import collections
import contextlib
import dataclasses
import itertools
import operator
import os
from typing import (
    Tuple,
    Optional,
    Dict,
    TypeVar,
    Callable,
    List,
    Any,
    Union,
    Mapping,
    IO,
    Iterator,
    NotRequired,
    Annotated,
    Generic,
)

from debian.debian_support import DpkgArchTable

from debputy.builtin_manifest_rules import (
    builtin_path_metadata_rules,
)
from debputy.highlevel_manifest import (
    HighLevelManifest,
    PackageTransformationDefinition,
    MutableYAMLManifest,
    MANIFEST_YAML,
)
from debputy.maintscript_snippet import (
    DpkgMaintscriptHelperCommand,
    MaintscriptSnippet,
    STD_CONTROL_SCRIPTS,
    MaintscriptSnippetContainer,
)
from debputy.manifest_path_rules import ManifestPathRule
from debputy.packages import BinaryPackage
from debputy.path_matcher import MatchRuleType, ExactFileSystemPath, MatchRule
from debputy.substitution import Substitution
from debputy.transformation_rules import (
    TransformationRule,
    ExclusionTransformationRule,
    CreateDirectoryTransformationRule,
    CreateVirtualPathTransformationRule,
    MoveTransformationRule,
)
from debputy.util import (
    _normalize_path,
    PKGVERSION_REGEX,
    PKGNAME_REGEX,
    escape_shell,
    active_profiles_match,
)
from debputy.util import _warn, _info
from ._deb_options_profiles import DebOptionsAndProfiles
from ._manifest_constants import *
from .architecture_support import DpkgArchitectureBuildProcessValuesTable
from .installations import InstallRule
from .manifest_conditions import (
    ManifestCondition,
    ArchMatchManifestCondition,
    BuildProfileMatch,
)
from .manifest_parser.base_types import (
    DebputyParsedContentStandardConditional,
    FileSystemMode,
    ROOT_DEFINITION,
    FileSystemGroup,
    FileSystemOwner,
    DebputyParsedContent,
)
from .manifest_parser.declarative_parser import (
    DeclarativeInputParser,
    DebputyParseHint,
    parser_from_typed_dict,
)
from .manifest_parser.exceptions import ManifestParseException
from .manifest_parser.parser_data import ParserContextData
from .manifest_parser.util import AttributePath
from .packager_provided_files import detect_all_packager_provided_files
from .plugin.api.impl import PluginProvidedFeatureSet

try:
    from Levenshtein import distance
except ImportError:

    def _detect_possible_typo(
        _d,
        _key,
        _attribute_parent_path: AttributePath,
        required: bool,
    ) -> None:
        if required:
            _info(
                "Install python3-levenshtein to have debputy try to detect typos in the manifest."
            )

else:

    def _detect_possible_typo(
        d,
        key,
        _attribute_parent_path: AttributePath,
        _required: bool,
    ) -> None:
        k_len = len(key)
        for actual_key in d:
            if abs(k_len - len(actual_key)) > 2:
                continue
            d = distance(key, actual_key)
            if d > 2:
                continue
            path = _attribute_parent_path.path
            ref = f'at "{path}"' if path else "at the manifest root level"
            _warn(
                f'Possible typo: The key "{actual_key}" should probably have been "{key}" {ref}'
            )


PF = TypeVar("PF")
TP = TypeVar("TP")


@dataclasses.dataclass(slots=True, frozen=True)
class DispatchingParser(Generic[PF, TP]):
    parser: DeclarativeInputParser[PF]
    handler: Callable[[str, PF, AttributePath, ParserContextData], TP]


# TODO: Make this something plugins can hook into
_INSTALL_RULE_PARSERS: Dict[str, DispatchingParser[PF, InstallRule]] = {}
_DPKG_CONFFILE_MGNT_PARSERS: Dict[
    str, DispatchingParser[PF, DpkgMaintscriptHelperCommand]
] = {}


_DIP = DeclarativeInputParser[PF]
_DIP_HDL = Callable[[str, PF, AttributePath, ParserContextData], TP]


def _create_dispatchable_parser_hook(
    dispatch_table: Dict[str, DispatchingParser[PF, TP]],
) -> Callable[[_DIP], Callable[[_DIP_HDL], _DIP_HDL]]:
    def _annotation(parser: _DIP) -> Callable[[_DIP_HDL], _DIP_HDL]:
        def _func(func: _DIP_HDL) -> _DIP_HDL:
            parse_logic = DispatchingParser(parser, func)
            if parser.name in dispatch_table:
                raise ValueError(f'The key "{parser.name}" is already taken')
            dispatch_table[parser.name] = parse_logic
            for alias in parser.aliases:
                if alias in dispatch_table:
                    raise ValueError(f'The key "{alias}" is already taken')
                dispatch_table[alias] = parse_logic
            return func

        return _func

    return _annotation


install_rule_parser = _create_dispatchable_parser_hook(_INSTALL_RULE_PARSERS)
dpkg_conffile_parser = _create_dispatchable_parser_hook(_DPKG_CONFFILE_MGNT_PARSERS)


def _per_package_subst_variables(
    p: BinaryPackage,
    *,
    name: Optional[str] = None,
) -> Dict[str, str]:
    return {
        "PACKAGE": name if name is not None else p.name,
        "_DEBPUTY_PKG_ARCH": p.package_deb_architecture_variable("ARCH"),
        "_DEBPUTY_PKG_ARCHOS": p.package_deb_architecture_variable("ARCH_OS"),
    }


class ManifestRootRule(DebputyParsedContent):
    manifest_version: str
    installations: NotRequired[Optional[List[Any]]]
    packages: NotRequired[Dict[str, Any]]


class PathManifestRule(DebputyParsedContentStandardConditional):
    mode: NotRequired[FileSystemMode]
    owner: NotRequired[FileSystemOwner]
    group: NotRequired[FileSystemGroup]


class EnsureDirectoryRule(PathManifestRule):
    path: str


class CreateSymlinkRule(DebputyParsedContentStandardConditional):
    path: str
    target: str


class BinaryPackageRule(DebputyParsedContent):
    binary_version: NotRequired[str]
    transformations: NotRequired[List[Any]]
    create_symlinks: NotRequired[List[Any]]
    directories: NotRequired[List[Any]]
    path_metadata: NotRequired[Dict[str, Any]]
    conffile_management: NotRequired[List[Any]]


class DpkgConffileManagementRuleBase(DebputyParsedContent):
    prior_to_version: NotRequired[str]
    owning_package: NotRequired[str]


class DpkgRenameConffileRule(DpkgConffileManagementRuleBase):
    source: str
    target: str


class DpkgRemoveConffileRule(DpkgConffileManagementRuleBase):
    path: str


class TransformationMoveRuleSpec(DebputyParsedContentStandardConditional):
    source: str
    target: str


class ParsedInstallRuleSourceFormat(DebputyParsedContentStandardConditional):
    sources: NotRequired[List[str]]
    source: Annotated[NotRequired[str], DebputyParseHint.target_attribute("sources")]
    into: Annotated[
        NotRequired[Union[str, List[str]]],
        DebputyParseHint.required_when_multi_binary(),
    ]
    exclude: NotRequired[Union[str, List[str]]]
    dest_dir: NotRequired[str]
    as_: NotRequired[
        Annotated[
            str,
            DebputyParseHint.conflicts_with_source_attributes(
                "sources", "dest_dir", "exclude"
            ),
            DebputyParseHint.target_attribute("install_as"),
        ]
    ]


class ParsedInstallRule(DebputyParsedContentStandardConditional):
    sources: List[str]
    into: NotRequired[List[BinaryPackage]]
    exclude: NotRequired[List[str]]
    dest_dir: NotRequired[str]
    install_as: NotRequired[str]


class ParsedInstallDocOrExamplesRuleSourceFormat(
    DebputyParsedContentStandardConditional
):
    sources: NotRequired[List[str]]
    source: Annotated[NotRequired[str], DebputyParseHint.target_attribute("sources")]
    into: Annotated[
        NotRequired[Union[str, List[str]]],
        DebputyParseHint.required_when_multi_binary(),
    ]
    exclude: NotRequired[Union[str, List[str]]]


class ParsedInstallDocOrExamplesRule(DebputyParsedContentStandardConditional):
    sources: List[str]
    into: NotRequired[List[BinaryPackage]]
    exclude: NotRequired[List[str]]


class ParsedInstallExcludeRuleSourceFormat(DebputyParsedContentStandardConditional):
    paths: NotRequired[List[str]]
    path: Annotated[NotRequired[str], DebputyParseHint.target_attribute("paths")]


class ParsedInstallExcludeRule(DebputyParsedContentStandardConditional):
    paths: List[str]


_INSTALL_PARSER = parser_from_typed_dict(
    MK_INSTALLATIONS_INSTALL,
    ParsedInstallRule,
    source_content=ParsedInstallRuleSourceFormat,
)
_INSTALL_DOCS_PARSER = parser_from_typed_dict(
    MK_INSTALLATIONS_INSTALL_DOCS,
    ParsedInstallDocOrExamplesRule,
    source_content=ParsedInstallDocOrExamplesRuleSourceFormat,
    aliases=[
        MK_INSTALLATIONS_INSTALL_EXAMPLES,
        "install-doc",
        "install-example",
    ],
)
_INSTALL_EXCLUDE_PARSER = parser_from_typed_dict(
    "exclude",
    ParsedInstallExcludeRule,
    source_content=ParsedInstallExcludeRuleSourceFormat,
)
_PATH_METADATA_RULE_PARSER = parser_from_typed_dict(
    "path-metadata (value)", PathManifestRule
)
_ENSURE_DIR_RULE_PARSER = parser_from_typed_dict(
    "directories (value)", EnsureDirectoryRule
)
_CREATE_SYMLINK_RULE_PARSER = parser_from_typed_dict(
    "create-symlink (value)", CreateSymlinkRule
)
_BINARY_PACKAGE_RULE_PARSER = parser_from_typed_dict(
    "binary-package-rule (value)", BinaryPackageRule
)
_DPKG_REMOVE_CONFFILE_PARSER = parser_from_typed_dict(
    "remove",
    DpkgRemoveConffileRule,
)
_DPKG_RENAME_CONFFILE_PARSER = parser_from_typed_dict(
    "rename",
    DpkgRenameConffileRule,
)
_TRANSFORMATION_MOVE = parser_from_typed_dict(
    "move",
    TransformationMoveRuleSpec,
)
_MANIFEST_ROOT_PARSER = parser_from_typed_dict(
    "root-parser", ManifestRootRule, allow_optional=True
)


# TODO: Should be configurable somehow
_SEARCH_DIRS = ["debian/tmp", "."]


@install_rule_parser(_INSTALL_PARSER)
def _install_rule_handler(
    _name: str,
    parsed_data: ParsedInstallRule,
    path: AttributePath,
    context: ParserContextData,
) -> InstallRule:
    sources_path = path["sources"]
    sources = [_parse_filename(p, sources_path) for p in parsed_data["sources"]]
    install_as = parsed_data.get("install_as")
    into = parsed_data.get("into")
    dest_dir = parsed_data.get("dest_dir")
    condition = parsed_data.get("when")
    if not into:
        into = [context.single_binary_package()]
    into = frozenset(into)
    excludes_raw = parsed_data.get("exclude")
    excludes = []
    if excludes_raw:
        excl_path = path["exclude"]
        excludes = [_parse_filename(p, excl_path) for p in excludes_raw]
    if install_as is not None:
        assert len(sources) == 1
        assert dest_dir is None
        return InstallRule.install_as(
            sources[0],
            install_as,
            into,
            _SEARCH_DIRS,
            path.path,
            condition,
        )
    return InstallRule.install_dest(
        sources,
        dest_dir,
        into,
        excludes,
        _SEARCH_DIRS,
        path.path,
        condition,
    )


@install_rule_parser(_INSTALL_DOCS_PARSER)
def _install_docs_or_examples_rule_handler(
    name: str,
    parsed_data: ParsedInstallDocOrExamplesRule,
    path: AttributePath,
    context: ParserContextData,
) -> InstallRule:
    sources = parsed_data["sources"]
    into = parsed_data.get("into")
    if not into:
        into = [context.single_binary_package()]
    condition = parsed_data.get("when")
    into = frozenset(into)
    excludes_raw = parsed_data.get("exclude")
    excludes = []
    if excludes_raw:
        excl_path = path["exclude"]
        excludes = [_parse_filename(p, excl_path) for p in excludes_raw]
    if name in (MK_INSTALLATIONS_INSTALL_EXAMPLES, "install-example"):
        install_rule = InstallRule.install_examples
    else:
        install_rule = InstallRule.install_doc
    return install_rule(
        sources,
        into,
        excludes,
        _SEARCH_DIRS,
        path.path,
        condition,
    )


@install_rule_parser(_INSTALL_EXCLUDE_PARSER)
def _install_exclude_rule_handler(
    _name: str,
    parsed_data: ParsedInstallExcludeRule,
    path: AttributePath,
    _context: ParserContextData,
) -> InstallRule:
    attr_path = path["paths"]
    paths = [_parse_filename(p, attr_path) for p in parsed_data["paths"]]
    condition = parsed_data.get("when")
    return InstallRule.exclude_paths(
        paths,
        _SEARCH_DIRS,
        path.path,
        condition,
    )


def _parse_filename(
    filename: str,
    attribute_path: AttributePath,
    *,
    allow_directories: bool = True,
) -> str:
    try:
        normalized_path = _normalize_path(filename, with_prefix=False)
    except ValueError as e:
        raise ManifestParseException(
            f'Error parsing the path "{filename}" defined in {attribute_path.path}: {e.args[0]}'
        ) from None
    if not allow_directories and filename.endswith("/"):
        raise ManifestParseException(
            f'The path "{filename}" in {attribute_path.path} ends with "/" implying it is a directory,'
            f" but this feature can only be used for files"
        )
    if normalized_path == ".":
        raise ManifestParseException(
            f'The path "{filename}" in {attribute_path.path} looks like the root directory,'
            f" but this feature does not allow the root directory here."
        )
    return normalized_path


def _parse_conffile_prior_version_and_owning_package(
    d: DpkgConffileManagementRuleBase,
    attribute_path: AttributePath,
) -> Tuple[Optional[str], Optional[str]]:
    prior_version = d.get("prior_to_version")
    owning_package = d.get("owning_package")

    if prior_version is not None and not PKGVERSION_REGEX.match(prior_version):
        p = attribute_path["prior_to_version"]
        raise ManifestParseException(
            f"The {MK_CONFFILE_MANAGEMENT_X_PRIOR_TO_VERSION} parameter in {p.path} must be a"
            r" valid package version (i.e., match (?:\d+:)?\d[0-9A-Za-z.+:~]*(?:-[0-9A-Za-z.+:~]+)*)."
        )

    if owning_package is not None and not PKGNAME_REGEX.match(owning_package):
        p = attribute_path["owning_package"]
        raise ManifestParseException(
            f"The {MK_CONFFILE_MANAGEMENT_X_OWNING_PACKAGE} parameter in {p.path} must be a valid"
            f" package name (i.e., match {PKGNAME_REGEX.pattern})."
        )

    return prior_version, owning_package


@dpkg_conffile_parser(_DPKG_RENAME_CONFFILE_PARSER)
def _dpkg_conffile_rename(
    _name: str,
    parsed_data: DpkgRenameConffileRule,
    path: AttributePath,
    _context: ParserContextData,
) -> DpkgMaintscriptHelperCommand:
    source_file = parsed_data["source"]
    target_file = parsed_data["target"]
    normalized_source = _parse_filename(
        source_file,
        path["source"],
        allow_directories=False,
    )
    path.path_hint = source_file

    normalized_target = _parse_filename(
        target_file,
        path["target"],
        allow_directories=False,
    )
    normalized_source = "/" + normalized_source
    normalized_target = "/" + normalized_target

    if normalized_source == normalized_target:
        raise ManifestParseException(
            f"Invalid rename defined in {path.path}: The source and target path are the same!"
        )

    version, owning_package = _parse_conffile_prior_version_and_owning_package(
        parsed_data, path
    )
    return DpkgMaintscriptHelperCommand.mv_conffile(
        path,
        normalized_source,
        normalized_target,
        version,
        owning_package,
    )


@dpkg_conffile_parser(_DPKG_REMOVE_CONFFILE_PARSER)
def _dpkg_conffile_remove(
    _name: str,
    parsed_data: DpkgRemoveConffileRule,
    path: AttributePath,
    _context: ParserContextData,
) -> DpkgMaintscriptHelperCommand:
    source_file = parsed_data["path"]
    normalized_source = _parse_filename(
        source_file,
        path["path"],
        allow_directories=False,
    )
    path.path_hint = source_file

    normalized_source = "/" + normalized_source

    version, owning_package = _parse_conffile_prior_version_and_owning_package(
        parsed_data, path
    )
    return DpkgMaintscriptHelperCommand.rm_conffile(
        path,
        normalized_source,
        version,
        owning_package,
    )


class HighLevelManifestParser(ParserContextData):
    def __init__(
        self,
        manifest_path: str,
        binary_packages: Mapping[str, BinaryPackage],
        substitution: Substitution,
        dpkg_architecture_variables: DpkgArchitectureBuildProcessValuesTable,
        dpkg_arch_query_table: DpkgArchTable,
        build_env: DebOptionsAndProfiles,
        plugin_provided_feature_set: PluginProvidedFeatureSet,
        *,
        # Available for testing purposes only
        debian_dir: str = "./debian",
    ):
        self.manifest_path = manifest_path
        self._binary_packages = binary_packages
        self._mutable_yaml_manifest: Optional[MutableYAMLManifest] = None
        self._substitution = substitution
        self._dpkg_architecture_variables = dpkg_architecture_variables
        self._dpkg_arch_query_table = dpkg_arch_query_table
        self._build_env = build_env
        self._substitution_stack = [self._substitution]
        self._package_state_stack = []
        self._plugin_provided_feature_set = plugin_provided_feature_set
        self._reasons_for_assembly_with_dpkg_deb_requiring_root: Dict[str, str] = {}
        self._debian_dir = debian_dir  # TODO: Make this an iterator or something else that is easier to mock/stub

        self._all_package_states = {
            n: PackageTransformationDefinition(
                binary_package=p,
                substitution=substitution.with_extra_substitutions(
                    **_per_package_subst_variables(p)
                ),
                is_auto_generated_package=False,
                maintscript_snippets=collections.defaultdict(
                    MaintscriptSnippetContainer
                ),
            )
            for n, p in binary_packages.items()
        }
        for n, p in binary_packages.items():
            dbgsym_name = f"{n}-dbgsym"
            if dbgsym_name in self._all_package_states:
                continue
            self._all_package_states[dbgsym_name] = PackageTransformationDefinition(
                binary_package=p,
                substitution=substitution.with_extra_substitutions(
                    **_per_package_subst_variables(p, name=dbgsym_name)
                ),
                is_auto_generated_package=True,
                maintscript_snippets=collections.defaultdict(
                    MaintscriptSnippetContainer
                ),
            )

        self._install_rules: Optional[List[InstallRule]] = None

        self._owner_cache = {ROOT_DEFINITION.entity_name: ROOT_DEFINITION}
        self._group_cache = {ROOT_DEFINITION.entity_name: ROOT_DEFINITION}
        self._uid_cache = {v.entity_id: v for v in self._owner_cache.values()}
        self._gid_cache = {v.entity_id: v for v in self._group_cache.values()}
        self._ownership_caches_loaded = False
        self._used = False

    @property
    def binary_packages(self) -> Mapping[str, BinaryPackage]:
        return self._binary_packages

    @property
    def _package_states(self) -> Mapping[str, PackageTransformationDefinition]:
        return self._all_package_states

    def build_manifest(self) -> HighLevelManifest:
        if self._used:
            raise TypeError("build_manifest can only be called once!")
        self._used = True
        if isinstance(self, YAMLManifestParser) and self._mutable_yaml_manifest is None:
            self._mutable_yaml_manifest = MutableYAMLManifest.empty_manifest()
        with os.scandir(self._debian_dir) as scan_dir_iter:
            all_packager_provided_files = detect_all_packager_provided_files(
                self._plugin_provided_feature_set.packager_provided_files,
                scan_dir_iter,
                self.binary_packages,
                lambda n: self._package_states[n].substitution,
            )
        for package in self._package_states:
            with self.binary_package_context(package) as context:
                if not context.is_auto_generated_package:
                    pff_result = all_packager_provided_files[package]
                    for manifest_info in pff_result.auto_installable:
                        self._add_entry(manifest_info)
                    context.reserved_packager_provided_files.update(
                        pff_result.reserved_only
                    )
                for manifest_info in builtin_path_metadata_rules(
                    self._dpkg_architecture_variables,
                    context.binary_package,
                    context.substitution,
                ):
                    self._add_entry(manifest_info)
                if context.ensure_directories:
                    context.transformations.append(
                        CreateDirectoryTransformationRule(
                            context.ensure_directories.values()
                        )
                    )
                if context.create_virtual_paths:
                    context.transformations.append(
                        CreateVirtualPathTransformationRule(
                            context.create_virtual_paths.values()
                        )
                    )
                self._transform_dpkg_maintscript_helpers_to_snippets()

        return HighLevelManifest(
            self.manifest_path,
            self._mutable_yaml_manifest,
            self._install_rules,
            self.binary_packages,
            self.substitution,
            self._package_states,
            self._dpkg_architecture_variables,
            self._dpkg_arch_query_table,
            self._build_env,
            self._reasons_for_assembly_with_dpkg_deb_requiring_root,
            self._plugin_provided_feature_set,
        )

    @contextlib.contextmanager
    def binary_package_context(
        self, package_name: str
    ) -> Iterator[PackageTransformationDefinition]:
        if package_name not in self._package_states:
            self._error(
                f'The package "{package_name}" is not present in the debian/control file (could not find'
                f' "Package: {package_name}" in a binary paragraph) nor is it a -dbgsym package for one'
                " for a package in debian/control."
            )
        package_state = self._package_states[package_name]
        self._substitution_stack.append(package_state.substitution)
        self._package_state_stack.append(package_state)
        ss_len = len(self._substitution_stack)
        ps_len = len(self._package_state_stack)
        yield package_state
        if ss_len != len(self._substitution_stack) or ps_len != len(
            self._package_state_stack
        ):
            raise RuntimeError("Internal error: Unbalanced stack manipulation detected")
        self._substitution_stack.pop()
        self._package_state_stack.pop()

    @property
    def substitution(self) -> Substitution:
        return self._substitution_stack[-1]

    @property
    def current_binary_package_state(self) -> PackageTransformationDefinition:
        if not self._package_state_stack:
            raise RuntimeError("Invalid state: Not in a binary package context")
        return self._package_state_stack[-1]

    def _transform_dpkg_maintscript_helpers_to_snippets(self) -> None:
        package_state = self.current_binary_package_state
        for dmh in package_state.dpkg_maintscript_helper_snippets:
            snippet = MaintscriptSnippet(
                definition_source=dmh.definition_source,
                snippet=f'dpkg-maintscript-helper {escape_shell(*dmh.cmdline)} -- "$@"\n',
            )
            for script in STD_CONTROL_SCRIPTS:
                package_state.maintscript_snippets[script].append(snippet)

    def normalize_path(
        self,
        path: str,
        definition_source: AttributePath,
        *,
        allow_root_dir_match: bool = False,
    ) -> ExactFileSystemPath:
        try:
            normalized = _normalize_path(path)
        except ValueError:
            self._error(
                f'The path "{path}" provided in {definition_source.path} should be relative to the root of the'
                ' package and not use any ".." or "." segments.'
            )
        if normalized == "." and not allow_root_dir_match:
            self._error(
                "Manifests must not change the root directory of the deb file.  Please correct"
                f' "{definition_source.path}" (path: "{path}) in {self.manifest_path}'
            )
        return ExactFileSystemPath(
            self.substitution.substitute(normalized, definition_source.path)
        )

    def parse_path_or_glob(
        self,
        path_or_glob: str,
        definition_source: AttributePath,
    ) -> MatchRule:
        match_rule = MatchRule.from_path_or_glob(
            path_or_glob, definition_source.path, substitution=self.substitution
        )
        # NB: "." and "/" will be translated to MATCH_ANYTHING by MatchRule.from_path_or_glob,
        # so there is no need to check for an exact match on "." like in normalize_path.
        if match_rule.rule_type == MatchRuleType.MATCH_ANYTHING:
            self._error(
                f'The chosen match rule "{path_or_glob}" matches everything (including the deb root directory).'
                f' Please correct "{definition_source.path}" (path: "{path_or_glob}) in {self.manifest_path} to'
                f' something that matches "less" than everything.'
            )
        return match_rule

    def parse_manifest(self) -> HighLevelManifest:
        raise NotImplementedError

    def _check_existing(
        self,
        match_rule: MatchRule,
        path_info: ManifestPathRule,
    ) -> bool:
        package_state = self.current_binary_package_state
        if match_rule in package_state.path_specs:
            other_path_spec = package_state.path_specs[match_rule]
            if (
                path_info.is_builtin_definition
                and not other_path_spec.is_builtin_definition
            ):
                return False
            self._error(
                f'The path "{match_rule.describe_match_short()}" has conflicting definitions (or is defined'
                f" twice) in manifest {self.manifest_path}. The two definitions are:\n"
                f" * {other_path_spec.definition_source}\n"
                f" * {path_info.definition_source}\n"
                "Please consolidate them into one definition."
            )
        if path_info.is_builtin_definition and match_rule.is_path_type_part_of_match:
            # if the user provided a rule that overlaps with a built-in rule, we also discard the built-in rule, if
            # has a path type restriction while the user provided does not.  Notably, users are currently unable to
            # provide a path type restricted rules.
            #
            # This is the logic that ensures that the user can write "remove *.la" and not be steam rolled by the
            # built-in "normalize permissions for *.la files" rule.
            shadowing_match_rule = match_rule.match_rule_without_path_type_match()
            return shadowing_match_rule not in package_state.path_specs
        return True

    def _add_entry(self, path_info: ManifestPathRule) -> None:
        package_state = self.current_binary_package_state
        match_rule = path_info.match_rule
        if not self._check_existing(match_rule, path_info):
            return

        package_name = package_state.binary_package.name

        if (
            package_name not in self._reasons_for_assembly_with_dpkg_deb_requiring_root
            and (path_info.uid != 0 or path_info.gid != 0)
        ):
            self._reasons_for_assembly_with_dpkg_deb_requiring_root[
                package_name
            ] = path_info.definition_source

        package_state.path_specs[match_rule] = path_info
        if path_info.rule_type.creates_path:
            assert isinstance(match_rule, ExactFileSystemPath)
            assert path_info.member_path not in package_state.create_virtual_paths
            package_state.create_virtual_paths[path_info.member_path] = path_info
            current_path_info = path_info
            parent_dir = ExactFileSystemPath(os.path.dirname(match_rule.path))
            assert parent_dir.path.startswith(
                "./"
            ), f'{parent_dir.path} should have started with "./" ({path_info.member_path})'
            while True:
                parent_path_info = package_state.path_specs.get(parent_dir)
                existed = False
                if parent_path_info is None:
                    parent_path_info = ManifestPathRule.directory(
                        path_info.definition_source,
                        parent_dir,
                        triggered_by=path_info,
                    )
                    package_state.path_specs[parent_dir] = parent_path_info
                    package_state.ensure_directories[parent_dir.path] = parent_path_info
                else:
                    existed = True
                parent_path_info.add_virtual_child(current_path_info)
                if existed or parent_dir.path == ".":
                    break
                current_path_info = parent_path_info
                parent_dir = ExactFileSystemPath(os.path.dirname(parent_dir.path))


class YAMLManifestParser(HighLevelManifestParser):
    def _optional_key(
        self,
        d,
        key,
        attribute_parent_path: AttributePath,
        expected_type=None,
        default_value=None,
    ):
        v = d.get(key)
        if v is None:
            _detect_possible_typo(d, key, attribute_parent_path, False)
            return default_value
        if expected_type is not None:
            return self._ensure_value_is_type(
                v, expected_type, key, attribute_parent_path
            )
        return v

    def _required_key(
        self,
        d,
        key,
        attribute_parent_path: AttributePath,
        expected_type=None,
        extra: Optional[Union[str, Callable[[], str]]] = None,
    ):
        v = d.get(key)
        if v is None:
            _detect_possible_typo(d, key, attribute_parent_path, True)
            if extra is not None:
                msg = extra if isinstance(extra, str) else extra()
                extra_info = " " + msg
            else:
                extra_info = ""
            self._error(
                f'Missing required key {key} at {attribute_parent_path.path} in manifest "{self.manifest_path}.'
                f"{extra_info}"
            )

        if expected_type is not None:
            return self._ensure_value_is_type(
                v, expected_type, key, attribute_parent_path
            )
        return v

    def _ensure_value_is_type(
        self,
        v,
        t,
        key: Union[str, int, AttributePath],
        attribute_parent_path: Optional[AttributePath],
    ):
        if v is None:
            return None
        if not isinstance(v, t):
            if isinstance(t, tuple):
                t_msg = "one of: " + ", ".join(x.__name__ for x in t)
            else:
                t_msg = f"a {t.__name__}"
            key_path = (
                key.path
                if isinstance(key, AttributePath)
                else attribute_parent_path[key]
            )
            self._error(
                f'The key {key_path.path} must be {t_msg} in manifest "{self.manifest_path}"'
            )
        return v

    def _parse_yaml_path_metadata(
        self,
        d: BinaryPackageRule,
        orig_definition_source: AttributePath,
    ) -> None:
        raw_path_specs = d.get("path_metadata", {})
        for path_spec, path_metadata in raw_path_specs.items():
            definition_source = orig_definition_source["path-metadata"]
            # Ensure path_spec is a str for the error m
            self._ensure_value_is_type(
                path_spec, str, f"{path_spec}", definition_source
            )
            if path_metadata is None:
                path_metadata = {}
            else:
                self._ensure_value_is_type(
                    path_metadata, dict, path_spec, definition_source
                )
            # Extra quotes is deliberate
            definition_source = definition_source[f'"{path_spec}"']
            parsed = _PATH_METADATA_RULE_PARSER.parse_input(
                path_metadata, definition_source, self
            )
            owner = parsed.get("owner", ROOT_DEFINITION)
            group = parsed.get("group", ROOT_DEFINITION)

            path_info = ManifestPathRule.ensure_metadata(
                definition_source,
                self.parse_path_or_glob(path_spec, definition_source),
                mode=parsed.get("mode"),
                uid=owner.entity_id,
                uname=owner.entity_name,
                gid=group.entity_id,
                gname=group.entity_name,
                condition=parsed.get("when"),
            )
            self._add_entry(path_info)

    def _parse_yaml_create_symlinks(
        self,
        d: BinaryPackageRule,
        orig_definition_source: AttributePath,
    ) -> None:
        raw_path_specs = d.get("create_symlinks", [])
        package_state = self.current_binary_package_state
        for no, symlink_spec in enumerate(raw_path_specs):
            definition_source = orig_definition_source[MK_CREATE_SYMLINKS][no]
            parsed = _CREATE_SYMLINK_RULE_PARSER.parse_input(
                symlink_spec, definition_source, self
            )
            link_path = parsed["path"]
            link_target = parsed["target"]
            condition = parsed.get("when")
            # Include the name in the source as it is easier for people to find than counting their way through
            # the list.
            definition_source.path_hint = link_path
            normalized_link_path = self.normalize_path(link_path, definition_source)
            path_info = ManifestPathRule.symlink(
                definition_source,
                normalized_link_path,
                link_target=self.substitution.substitute(
                    link_target, definition_source.path
                ),
                condition=condition,
            )
            directory = os.path.dirname(path_info.member_path)
            if directory not in package_state.ensure_directories:
                package_state.ensure_directories[
                    directory
                ] = ManifestPathRule.directory(
                    definition_source,
                    ExactFileSystemPath(directory),
                    triggered_by=path_info,
                )
            self._add_entry(path_info)

    def _unpack_single_key_dict(self, value, definition_source: AttributePath) -> str:
        self._ensure_value_is_type(value, dict, definition_source, None)
        if len(value) != 1:
            self._error(
                f'The mapping "{definition_source}" had two keys, but it should only have one top level key.'
                " Maybe you are missing a list marker behind the second key or some indentation.  The keys"
                f' are: {", ".join(value)}'
            )
        return next(iter(value))

    def _parse_yaml_directories(
        self,
        d: BinaryPackageRule,
        orig_definition_source: AttributePath,
    ) -> None:
        raw_path_specs = d.get("directories", [])
        directories = []
        package_state = self.current_binary_package_state
        parent_path = orig_definition_source["directories"]
        for no, dir_def in enumerate(raw_path_specs):
            self._ensure_value_is_type(dir_def, (str, dict), no, parent_path)
            definition_source = parent_path[no]
            condition = None
            mode = None
            if isinstance(dir_def, str):
                dir_path = dir_def
                owner = ROOT_DEFINITION
                group = ROOT_DEFINITION
            else:
                assert isinstance(dir_def, dict)
                dir_data = dir_def
                parsed = _ENSURE_DIR_RULE_PARSER.parse_input(
                    dir_data, definition_source, self
                )
                dir_path = parsed["path"]
                mode = parsed.get("mode")
                owner = parsed.get("owner", ROOT_DEFINITION)
                group = parsed.get("group", ROOT_DEFINITION)
                condition = parsed.get("when")

            # Include the name in the source as it is easier for people to find than counting their way through
            # the list.
            definition_source.path_hint = dir_path
            normalized_dir_path = self.normalize_path(dir_path, definition_source)
            path_info = ManifestPathRule.directory(
                definition_source,
                normalized_dir_path,
                mode=mode,
                uid=owner.entity_id,
                uname=owner.entity_name,
                gid=group.entity_id,
                gname=group.entity_name,
                condition=condition,
            )
            directories.append(path_info)

        # Insert them in sort order as that will ensure the implicit path match the "shortest" rule rather than the
        # "first" rule.  It also saves us form having to do "conflict" resolution between implicit and explicit
        # directory rules.
        for path_info in sorted(directories, key=operator.attrgetter("member_path")):
            self._add_entry(path_info)
            assert path_info.member_path not in package_state.ensure_directories
            package_state.ensure_directories[path_info.member_path] = path_info

    def _parse_transformation_exclusion(
        self,
        d,
        key: str,
        definition_source: AttributePath,
    ) -> TransformationRule:
        exclusion = self._required_key(d, key, definition_source, expected_type=str)
        definition_source = definition_source[key]
        definition_source.path_hint = exclusion
        return ExclusionTransformationRule(
            self.parse_path_or_glob(exclusion, definition_source),
            definition_source,
        )

    def _parse_transformation_move(
        self,
        d,
        key: str,
        definition_source: AttributePath,
    ) -> TransformationRule:
        move_def = self._required_key(d, key, definition_source, expected_type=dict)
        definition_source = definition_source[key]
        parsed_data = _TRANSFORMATION_MOVE.parse_input(
            move_def, definition_source, self
        )
        source = parsed_data["source"]
        target = parsed_data["target"]
        condition = parsed_data.get("when")
        definition_source.path_hint = source
        source_match = self.parse_path_or_glob(source, definition_source["source"])
        target_path = self.normalize_path(target, definition_source["target"]).path
        if (
            isinstance(source_match, ExactFileSystemPath)
            and source_match.path == target_path
        ):
            self._error(
                f"The transformation rule {definition_source} requests a move of {source_match} to"
                f" {target_path}, which is the same path"
            )
        return MoveTransformationRule(
            source_match,
            target_path,
            target.endswith("/"),
            definition_source,
            condition,
        )

    def _parse_yaml_condition_not(
        self,
        d,
        key: str,
        orig_definition_source: AttributePath,
    ) -> ManifestCondition:
        not_def = self._required_key(d, key, orig_definition_source, expected_type=dict)
        condition = self.parse_manifest_condition(not_def, orig_definition_source[key])
        return condition.negated()

    def _parse_yaml_condition_group(
        self,
        d,
        key: str,
        orig_definition_source: AttributePath,
    ) -> ManifestCondition:
        group_def = self._required_key(
            d, key, orig_definition_source, expected_type=list
        )
        definition_source = orig_definition_source[key]
        conditions = [
            self.parse_manifest_condition(cd, definition_source[i])
            for i, cd in enumerate(group_def)
        ]
        if len(conditions) < 2:
            self._error(
                f"Invalid condition at {definition_source.path}, the {key} requires at least 2 conditions."
            )

        if key == "any-of":
            return ManifestCondition.any_of(conditions)
        assert key == "all-of"
        return ManifestCondition.all_of(conditions)

    def _parse_yaml_condition_arch_matches(
        self,
        d,
        key: str,
        orig_definition_source: AttributePath,
    ) -> ManifestCondition:
        arch_matches_raw = self._required_key(
            d,
            key,
            orig_definition_source,
            expected_type=str,
        )
        package_state = self.current_binary_package_state
        arch_matches_as_list = arch_matches_raw.split()
        dpkg_arch_query = self._dpkg_arch_query_table

        if not arch_matches_as_list:
            self._error(
                f"The condition at {orig_definition_source.path} must not be empty"
            )
        if package_state.binary_package.is_arch_all:
            if arch_matches_as_list[0].startswith("["):
                arch_matches_as_list[0] = arch_matches_as_list[0][1:]
            if arch_matches_as_list[-1].endswith("]"):
                arch_matches_as_list[-1] = arch_matches_as_list[-1][:-1]
            result = dpkg_arch_query.architecture_is_concerned(
                "all", arch_matches_as_list
            )
            self._error(
                f"The package architecture restriction at {orig_definition_source.path} is applied to the"
                f' "Architecture: all" package {package_state.binary_package.name}, which does not make sense'
                f" as the condition will always resolves to `{str(result).lower()}`."
            )
        if arch_matches_as_list[0].startswith("[") or arch_matches_as_list[-1].endswith(
            "]"
        ):
            self._error(
                f"The architecture match at {orig_definition_source} must be defined without enclosing it with "
                '"[" or/and "]" brackets'
            )
        # Can we check arch list for typos.  If we do, it must be tight in how close matches it does.
        # Consider "arm" vs. "armel" (edit distance 2, but both are valid).  Likewise, names often
        # include a bit indicator "foo", "foo32", "foo64" - all of these have an edit distance of 2
        # of each other.
        return ArchMatchManifestCondition(arch_matches_as_list)

    def _parse_yaml_condition_build_profiles_matches(
        self,
        d,
        key: str,
        orig_definition_source: AttributePath,
    ) -> ManifestCondition:
        build_profile_spec = self._required_key(
            d,
            key,
            orig_definition_source,
            expected_type=str,
        )
        build_profile_spec = build_profile_spec.strip()
        if not build_profile_spec:
            self._error(
                f"The condition at {orig_definition_source.path} must not be empty"
            )
        try:
            active_profiles_match(build_profile_spec, frozenset())
        except ValueError as e:
            self._error(
                f"Could not parse the build specification at {orig_definition_source.path}: {e.args[0]}"
            )
        return BuildProfileMatch(build_profile_spec)

    def parse_manifest_condition(
        self,
        d: Union[str, Mapping[str, Any]],
        orig_definition_source: AttributePath,
    ) -> ManifestCondition:
        supported_mapping_conditions: Dict[
            str, Callable[[Mapping, str, AttributePath], ManifestCondition]
        ] = {
            "not": self._parse_yaml_condition_not,
            MK_CONDITION_ARCH_MATCHES: self._parse_yaml_condition_arch_matches,
            MK_CONDITION_BUILD_PROFILES_MATCHES: self._parse_yaml_condition_build_profiles_matches,
            "any-of": self._parse_yaml_condition_group,
            "all-of": self._parse_yaml_condition_group,
        }
        supported_str_conditions: Dict[str, ManifestCondition] = {
            "cross-compiling": ManifestCondition.is_cross_building(),
            "can-execute-compiled-binaries": ManifestCondition.can_execute_compiled_binaries(),
            "run-build-time-tests": ManifestCondition.run_build_time_tests(),
        }
        condition = None
        if isinstance(d, str):
            condition_key = d
            condition = supported_str_conditions.get(condition_key)
        else:
            condition_key = self._unpack_single_key_dict(d, orig_definition_source)

            handler = supported_mapping_conditions.get(condition_key)
            if handler is not None:
                condition = handler(d, condition_key, orig_definition_source)

        if condition is None:
            if condition_key in supported_mapping_conditions:
                assert not isinstance(d, dict)
                self._error(
                    f"Invalid definition at {orig_definition_source.path}.  The condition {condition_key} must be a "
                    "mapping."
                )
            elif condition_key in supported_str_conditions:
                self._error(
                    f"Invalid definition at {orig_definition_source.path}.  The condition {condition_key} must be a "
                    "string."
                )
            dict_options = (f"{k} (mapping)" for k in supported_mapping_conditions)
            string_options = (f"{k} (string)" for k in supported_str_conditions)
            valid_options = ", ".join(
                sorted(itertools.chain(dict_options, string_options))
            )
            self._error(
                f'Unknown or unsupported condition "{condition_key}" at {orig_definition_source.path}.'
                f" Valid conditions are: {valid_options}"
            )

        return condition

    def _parse_yaml_installations(
        self,
        d: ManifestRootRule,
        orig_definition_source: AttributePath,
    ) -> None:
        raw_installations = d.get("installations")

        if raw_installations is None and MK_INSTALLATIONS not in d:
            # Explicitly empty vs. implicitly empty matters in corner cases, so preserve this
            # fact.
            return

        self._install_rules: List[InstallRule] = []
        parent_path = orig_definition_source[MK_INSTALLATIONS]

        for no, raw_installation in enumerate(raw_installations):
            definition_source = parent_path[no]
            installation_key = self._unpack_single_key_dict(
                raw_installation,
                definition_source,
            )
            handler = _INSTALL_RULE_PARSERS.get(installation_key)
            if handler is not None:
                entry_path = definition_source[installation_key]
                parsed_data = handler.parser.parse_input(
                    raw_installation[installation_key], entry_path, self
                )
                installation = handler.handler(
                    installation_key, parsed_data, entry_path, self
                )
                self._install_rules.append(installation)
            else:
                self._error(
                    f'Unknown or unsupported install rule "{installation_key}" at {definition_source.path}.'
                    " Installations must have one of the following keys:"
                    f' {", ".join(sorted(_INSTALL_RULE_PARSERS))}'
                )

    def _parse_yaml_transformations(
        self,
        d: BinaryPackageRule,
        transformations: List[TransformationRule],
        orig_definition_source: AttributePath,
    ) -> None:
        raw_transformations = d.get("transformations", [])
        supported_transformations: Dict[
            str, Callable[[Dict, str, AttributePath], TransformationRule]
        ] = {
            "exclude": self._parse_transformation_exclusion,
            "move": self._parse_transformation_move,
        }

        parent_path = orig_definition_source["transformations"]

        for no, raw_transformation in enumerate(raw_transformations):
            definition_source = parent_path[no]
            transformation_key = self._unpack_single_key_dict(
                raw_transformation,
                definition_source,
            )
            handler = supported_transformations.get(transformation_key)
            if handler is not None:
                transformation = handler(
                    raw_transformation,
                    transformation_key,
                    definition_source,
                )
                transformations.append(transformation)
            else:
                self._error(
                    f'Unknown or unsupported transformation "{transformation_key}" at {definition_source}.'
                    " Transformations must have one of the following keys:"
                    f' {", ".join(sorted(supported_transformations))}'
                )

    def _parse_yaml_conffile_management(
        self,
        d: BinaryPackageRule,
        definition_source: AttributePath,
    ) -> None:
        entries = d.get("conffile_management", [])
        package_state = self.current_binary_package_state

        parent_path = definition_source[MK_CONFFILE_MANAGEMENT]

        for no, entry in enumerate(entries):
            path_definition_source = parent_path[no]
            key = self._unpack_single_key_dict(entry, path_definition_source)
            handler = _DPKG_CONFFILE_MGNT_PARSERS.get(key)
            if handler is not None:
                entry_path = path_definition_source[key]
                parsed_data = handler.parser.parse_input(entry[key], entry_path, self)
                dpkg_script_snippet = handler.handler(
                    key, parsed_data, entry_path, self
                )
                package_state.dpkg_maintscript_helper_snippets.append(
                    dpkg_script_snippet
                )
            else:
                self._error(
                    f'Unknown or unsupported conffile management "{key}" at {path_definition_source.path}.'
                    " Conffile management must have one of the following keys:"
                    f' {", ".join(sorted(entry))}'
                )

    def from_yaml_dict(self, yaml_data):
        attribute_path = AttributePath.root_path()
        parsed_data = _MANIFEST_ROOT_PARSER.parse_input(yaml_data, attribute_path, self)
        manifest_version = parsed_data["manifest_version"]
        if manifest_version not in SUPPORTED_MANIFEST_VERSIONS:
            self._error(
                "Unsupported manifest-version.  This implementation supports the following versions:"
                f' {", ".join(repr(v) for v in SUPPORTED_MANIFEST_VERSIONS)}"'
            )

        packages_dict = parsed_data.get("packages", {})
        self._parse_yaml_installations(parsed_data, attribute_path)
        packages_parent_path = attribute_path["packages"]
        for k, v in packages_dict.items():
            definition_source = packages_parent_path[k]
            with self.binary_package_context(k) as package_state:
                if package_state.is_auto_generated_package:
                    # Maybe lift (part) of this restriction.
                    self._error(
                        f'Cannot define rules for package "{k}" (at {definition_source.path}). It is an'
                        " auto-generated package."
                    )
                parsed = _BINARY_PACKAGE_RULE_PARSER.parse_input(
                    v, definition_source, self
                )

                binary_version = parsed.get("binary_version")
                if binary_version is not None:
                    package_state.binary_version = (
                        package_state.substitution.substitute(
                            binary_version,
                            definition_source["binary-version"],
                        )
                    )
                # Parse directories first.  It makes it easier for us to get the implicit rules to make sense
                # for people and saves us from doing it conflict resolution between implicit and explicit
                # directory rules.
                self._parse_yaml_directories(parsed, definition_source)
                self._parse_yaml_path_metadata(parsed, definition_source)
                self._parse_yaml_create_symlinks(parsed, definition_source)
                self._parse_yaml_transformations(
                    parsed,
                    package_state.transformations,
                    definition_source,
                )
                self._parse_yaml_conffile_management(parsed, definition_source)
        return self.build_manifest()

    def parse_manifest(self, *, fd: Union[IO[bytes], str] = None) -> HighLevelManifest:
        if fd is None:
            with open(self.manifest_path) as fd:
                data = MANIFEST_YAML.load(fd)
                self._mutable_yaml_manifest = MutableYAMLManifest(data)
                return self.from_yaml_dict(data)
        else:
            data = MANIFEST_YAML.load(fd)
            self._mutable_yaml_manifest = MutableYAMLManifest(data)
            return self.from_yaml_dict(data)


def parse_manifest(
    manifest_path: Optional[str],
    binary_packages: Mapping[str, BinaryPackage],
    substitution: Substitution,
    dpkg_architecture_variables: DpkgArchitectureBuildProcessValuesTable,
    dpkg_arch_query_table: DpkgArchTable,
    build_env: DebOptionsAndProfiles,
    debputy_plugin_feature_set,
) -> HighLevelManifest:
    if manifest_path is None:
        manifest_path = "debian/debputy.manifest"
        if not os.path.exists(manifest_path):
            return YAMLManifestParser(
                manifest_path,
                binary_packages,
                substitution,
                dpkg_architecture_variables,
                dpkg_arch_query_table,
                build_env,
                debputy_plugin_feature_set,
            ).build_manifest()
    return YAMLManifestParser(
        manifest_path,
        binary_packages,
        substitution,
        dpkg_architecture_variables,
        dpkg_arch_query_table,
        build_env,
        debputy_plugin_feature_set,
    ).parse_manifest()
