import dataclasses
import glob
import os.path
import re
import stat
from enum import IntEnum
from os import DirEntry
from typing import (
    List,
    Dict,
    FrozenSet,
    Callable,
    Union,
    Iterator,
    Tuple,
    Set,
    Sequence,
    Optional,
    Mapping,
    Iterable,
    TYPE_CHECKING,
)

from debputy.filesystem_scan import FSPath, fs_scan_dir, FSPathDirEntry
from debputy.manifest_conditions import (
    ConditionContext,
    ManifestCondition,
    _BUILD_DOCS_BDO,
)
from debputy.packages import BinaryPackage
from debputy.util import has_glob_magic, _normalize_path, _error, _warn

if TYPE_CHECKING:
    from debputy.highlevel_manifest import BinaryPackageInstallRuleContext


class InstallRuleError(RuntimeError):
    @property
    def message(self) -> str:
        return self.args[0]


class PathAlreadyInstalledOrExcludedError(InstallRuleError):
    @property
    def path(self) -> str:
        return self.args[0]

    @property
    def into(self) -> FrozenSet[BinaryPackage]:
        return self.args[1]

    @property
    def definition_source(self) -> str:
        return self.args[2]


class ExactPathMatchTwiceError(InstallRuleError):
    @property
    def path(self) -> str:
        return self.args[1]

    @property
    def into(self) -> BinaryPackage:
        return self.args[2]

    @property
    def definition_source(self) -> str:
        return self.args[3]


class NoMatchForInstallPatternError(InstallRuleError):
    @property
    def pattern(self) -> str:
        return self.args[1]

    @property
    def search_dirs(self) -> Sequence[str]:
        return self.args[2]

    @property
    def definition_source(self) -> str:
        return self.args[3]


@dataclasses.dataclass(slots=True, frozen=True)
class PathMatch:
    path: str
    search_dir: str
    fs_path: str
    stat_cache: os.stat_result
    is_exact_match: bool


class ExcludeState(IntEnum):
    UNCHECKED = 0
    NOT_EXCLUDED = 1
    EXCLUDED_BY_INLINE_RULE = 2
    EXCLUDED_BY_PLUGIN_PROVIDED_RULE = 3
    EXCLUDED_BY_MANIFEST_RULE = 4


def _debputy_exclude_pyc_files(_path: str, basename: str) -> bool:
    return basename == "__pycache__" or basename.endswith((".pyc", ".pyo"))


def _debputy_prune_la_files(_path: str, basename: str) -> bool:
    return basename.endswith(".la")


_BACKUP_FILES_RE = re.compile(
    "|".join(
        [
            # Common backup files
            r".*~",
            r".*[.](?:bak|orig)",
            # Editor backup/swap files
            r"[.]#.*",
            r"[.].*[.]sw.",
            # Other known stuff
            r"[.]shelf",
            r",,.*",  # "baz-style junk" (according to dpkg (Dpkg::Source::Package)
            r"DEADJOE",  # Joe's one line of immortality that just gets cargo cult'ed around ... just in case.
        ]
    )
)


def _debputy_prune_backup_files(_path: str, basename: str) -> bool:
    return bool(_BACKUP_FILES_RE.match(basename))


_VCS_PATHS = {
    ".arch-inventory",
    ".arch-ids",
    ".be",
    ".bzrbackup",
    ".bzrignore",
    ".bzrtags",
    ".cvsignore",
    ".hg",
    ".hgignore",
    ".hgtags",
    ".hgsigs",
    ".git",
    ".gitignore",
    ".gitattributes",
    ".gitmodules",
    ".gitreview",
    ".mailmap",
    ".mtn-ignore",
    ".svn",
    "{arch}",
    "CVS",
    "RCS",
    "_MTN",
    "_darcs",
}


def _debputy_prune_vcs_files(_path: str, basename: str) -> bool:
    return basename in _VCS_PATHS


def _debputy_prune_info_dir_file(path: str, _basename) -> bool:
    return path.endswith("usr/share/info/dir")


def _debputy_prune_binary_debian_dir(_path: str, basename: str) -> bool:
    return basename == "DEBIAN"


_PLUGIN_PROVIDED_EXCLUDES = [
    _debputy_exclude_pyc_files,
    _debputy_prune_la_files,
    _debputy_prune_backup_files,
    _debputy_prune_vcs_files,
    _debputy_prune_info_dir_file,
    _debputy_prune_binary_debian_dir,
]


def _run_plugin_provided_exclude_rules_on(path: str) -> bool:
    basename = os.path.basename(path)
    return any(c(path, basename) for c in _PLUGIN_PROVIDED_EXCLUDES)


def _resolve_inline_excludes(
    excludes: List[Tuple[str, bool]],
    match: PathMatch,
    dir_only_match: bool,
    definition_source: str,
) -> Mapping[str, str]:
    resolved = {}
    for e, e_is_glob in excludes:
        resolved_len = len(resolved)

        rule_resolved = {
            m.fs_path: e
            for m in _resolve_path(e, e_is_glob, [match.search_dir], dir_only_match)
        }
        resolved.update(rule_resolved)
        if len(resolved) == resolved_len:
            # TODO: Can we promote the warning to an error when there are "no mitigating" circumstances,
            #  so a regular build will detect this as superfluous?
            _warn(
                f"Resolving the exclude rule {e} defined at {definition_source} did not match any paths"
                f" in {match.search_dir}.  This might be a typo in the exclude rule or the path might not exist"
                " due to the DEB_BUILD_OPTIONS/DEB_BUILD_PROFILES"
            )
    return resolved


class SourcePathMatcher:
    def __init__(self) -> None:
        self._already_matched: Dict[
            str,
            Tuple[FrozenSet[BinaryPackage], str],
        ] = {}
        self._exact_match_request: Set[Tuple[str, str]] = set()
        self._excluded: Dict[str, ExcludeState] = {}

    def is_reserved(self, entry: DirEntry[str]) -> bool:
        path = entry.path
        if path in self._already_matched:
            return True
        result = self._excluded.get(path, ExcludeState.UNCHECKED)
        if result == ExcludeState.UNCHECKED:
            result = self._check_plugin_provided_exclude_state_for(path, ".")
        if result == ExcludeState.NOT_EXCLUDED:
            return False

        return True

    def exclude(self, path: str, *, inline_exclusion: bool = False) -> None:
        e = ExcludeState.EXCLUDED_BY_MANIFEST_RULE
        if inline_exclusion:
            e = ExcludeState.EXCLUDED_BY_INLINE_RULE
        self._excluded[path] = e

    def _check_plugin_provided_exclude_state_for(
        self, path: str, search_dir: str
    ) -> ExcludeState:
        cache_misses = []
        current_path = path
        while True:
            exclude_state = self._excluded.get(current_path, ExcludeState.UNCHECKED)
            if exclude_state != ExcludeState.UNCHECKED:
                verdict = exclude_state
                break
            cache_misses.append(current_path)
            if _run_plugin_provided_exclude_rules_on(current_path):
                verdict = ExcludeState.EXCLUDED_BY_PLUGIN_PROVIDED_RULE
                break
            # We cannot trust a "NOT_EXCLUDED" until we check its parent (the directory could
            # be excluded without the files in it triggering the rule).
            current_path = os.path.dirname(current_path)
            if not current_path or current_path == search_dir:
                verdict = ExcludeState.NOT_EXCLUDED
                break
        if cache_misses:
            for p in cache_misses:
                self._excluded[p] = verdict
        return verdict

    def may_match(
        self,
        match: PathMatch,
        *,
        is_exact_match: bool = False,
    ) -> Tuple[FrozenSet[BinaryPackage], bool]:
        m = self._already_matched.get(match.fs_path)
        if m:
            return m[0], False
        current_path = match.fs_path
        exclude_state = self._excluded.get(current_path, ExcludeState.UNCHECKED)

        if exclude_state == ExcludeState.UNCHECKED:
            exclude_state = self._check_plugin_provided_exclude_state_for(
                match.path, match.search_dir
            )

        assert exclude_state is not None

        is_excluded = exclude_state != ExcludeState.NOT_EXCLUDED
        if (
            is_exact_match
            and exclude_state == ExcludeState.EXCLUDED_BY_PLUGIN_PROVIDED_RULE
        ):
            is_excluded = False
        return frozenset(), is_excluded

    def reserve(
        self,
        path: str,
        reserved_by: FrozenSet[BinaryPackage],
        definition_source: str,
        *,
        is_exact_match: bool = False,
    ) -> None:
        assert _normalize_path(path, with_prefix=False) == path

        self._already_matched[path] = reserved_by, definition_source
        if not is_exact_match:
            return
        for pkg in reserved_by:
            m_key = (pkg.name, path)
            self._exact_match_request.add(m_key)
        try:
            del self._excluded[path]
        except KeyError:
            pass

    def detect_missing(self, search_dir: str) -> Iterator[os.DirEntry[str]]:
        stack = [e for e in fs_scan_dir(search_dir)]
        while stack:
            m = stack.pop()
            if m.is_dir(follow_symlinks=False):
                s_len = len(stack)
                stack.extend(de for de in fs_scan_dir(m.path))

                if s_len == len(stack) and not self.is_reserved(m):
                    # "Explicitly" empty dir
                    yield m
            elif not self.is_reserved(m):
                yield m

    def find_and_reserve_all_matches(
        self,
        norm_pattern: str,
        excludes: List[Tuple[str, bool]],
        search_dirs: Sequence[str],
        glob_expand: bool,
        dir_only_match: bool,
        reserved_by: FrozenSet[BinaryPackage],
        definition_source: str,
    ) -> Tuple[List[PathMatch], Tuple[int, ...]]:
        matched = []
        already_installed_paths = 0
        already_excluded_paths = 0
        self_excludes = 0
        used_self_excludes = set()
        excludes_resolved = None
        if excludes and not glob_expand:
            raise ValueError(f"Cannot use excludes without glob expansion")

        for match in _resolve_path(
            norm_pattern,
            glob_expand,
            search_dirs,
            dir_only_match,
        ):
            if excludes_resolved is None:
                excludes_resolved = _resolve_inline_excludes(
                    excludes, match, dir_only_match, definition_source
                )
            self_exclude = excludes_resolved.get(match.fs_path)
            if self_exclude is not None:
                used_self_excludes.add(self_exclude)
                self.exclude(match.fs_path, inline_exclusion=True)
                self_excludes += 1
                continue
            installed_into, excluded = self.may_match(
                match, is_exact_match=not glob_expand
            )
            if installed_into:
                if glob_expand:
                    already_installed_paths += 1
                    continue
                packages = ", ".join(p.name for p in installed_into)
                raise PathAlreadyInstalledOrExcludedError(
                    f'The "{match.path}" has been reserved by and installed into {packages}.'
                    f" The definition that triggered this issue is {definition_source}.",
                    match,
                    installed_into,
                    definition_source,
                )
            if excluded:
                if glob_expand:
                    already_excluded_paths += 1
                    continue
                raise PathAlreadyInstalledOrExcludedError(
                    f'The "{match.path}" has been excluded. If you want this path installed, move it'
                    f" above the exclusion rule that excluded it. The definition that triggered this"
                    f" issue is {definition_source}.",
                    match,
                    installed_into,
                    definition_source,
                )
            if not glob_expand:
                for pkg in reserved_by:
                    m_key = (pkg.name, match.path)
                    if m_key in self._exact_match_request:
                        raise ExactPathMatchTwiceError(
                            f'The path "{match.path}" (via exact match) has already been installed'
                            f" into {pkg.name}. The second installation triggered by {definition_source}",
                            match.path,
                            pkg,
                            definition_source,
                        )
                    self._exact_match_request.add(m_key)

            if reserved_by:
                self._already_matched[match.fs_path] = reserved_by, definition_source
            else:
                self.exclude(match.fs_path, inline_exclusion=False)
            matched.append(match)
        exclude_counts = already_installed_paths, already_excluded_paths, self_excludes
        return matched, exclude_counts


def _resolve_path(
    norm_pattern: str,
    glob_expand: bool,
    search_dirs: Iterable[str],
    dir_only_match: bool,
) -> Iterator[PathMatch]:
    if glob_expand:
        for sdir in search_dirs:
            matched = False
            for match in glob.glob(norm_pattern, root_dir=sdir):
                fs_path = os.path.join(sdir, match)
                if dir_only_match:
                    fs_path += "/"
                try:
                    st = os.lstat(fs_path)
                except FileNotFoundError:
                    continue
                else:
                    yield PathMatch(match, sdir, fs_path, st, False)
                    matched = True
            if matched:
                break
    else:
        for sdir in search_dirs:
            fs_path = os.path.join(sdir, norm_pattern)
            if dir_only_match:
                fs_path += "/"
            try:
                st = os.lstat(fs_path)
            except FileNotFoundError:
                continue
            else:
                yield PathMatch(norm_pattern, sdir, fs_path, st, True)
                break


def _resolve_matches(
    matches: List[PathMatch],
    dest_path: Union[str, Callable[[PathMatch], str]],
    dest_path_is_format: bool,
    install_context: "BinaryPackageInstallRuleContext",
) -> Iterator[Tuple[PathMatch, str]]:
    if callable(dest_path):
        if dest_path_is_format:
            raise ValueError(
                "dest_path_is_format cannot be True when dest_path is a Callable"
            )
        for match in matches:
            yield match, dest_path(match)
    else:
        for match in matches:
            if dest_path_is_format:
                yield match, dest_path.format(
                    basename=os.path.basename(match.path),
                    dirname=os.path.dirname(match.path),
                    package_name=install_context.binary_package.name,
                    doc_main_package_name=install_context.doc_main_package.name,
                )
            else:
                yield match, dest_path


class InstallRule:
    __slots__ = (
        "_already_matched",
        "_exact_match_request",
        "_into",
        "_exclude",
        "_search_dirs",
        "_condition",
        "_definition_source",
    )

    def __init__(
        self,
        into: FrozenSet[BinaryPackage],
        exclude: List[str],
        search_dirs: Sequence[str],
        condition: Optional[ManifestCondition],
        definition_source: str,
    ) -> None:
        self._into = into
        self._exclude = [(p, glob.has_magic(p)) for p in exclude]
        self._search_dirs = search_dirs
        self._condition = condition
        self._definition_source = definition_source
        # We heavily rely on normalization for "_already_matched"
        assert all(
            _normalize_path(s, with_prefix=False) == s for s in self._search_dirs
        )

    def _match_pattern(
        self,
        path_matcher: SourcePathMatcher,
        pattern: str,
        condition_context: ConditionContext,
    ) -> List[PathMatch]:
        norm_pattern = _normalize_path(pattern, with_prefix=False)
        is_glob = has_glob_magic(norm_pattern)
        (matched, exclude_counts) = path_matcher.find_and_reserve_all_matches(
            norm_pattern,
            self._exclude,
            self._search_dirs,
            is_glob,
            pattern.endswith("/"),
            self._into,
            self._definition_source,
        )

        already_installed_paths, already_excluded_paths, self_excludes = exclude_counts

        allow_empty_match = all(not p.should_be_acted_on for p in self._into)
        if self._condition is not None and not self._condition.evaluate(
            condition_context
        ):
            allow_empty_match = True

        if not matched and not allow_empty_match:
            combined_exclude = already_excluded_paths + self_excludes
            if combined_exclude and already_installed_paths:
                total_paths = already_excluded_paths + already_installed_paths
                msg = (
                    f"There were no matches for {pattern} in {self._search_dirs} after ignoring"
                    f" {total_paths} path(s) already been matched previously either by install or"
                    f" exclude rules. If you wanted to install some of these paths into multiple"
                    f" packages, please tweak the definition that installed them to install them"
                    f' into multiple packages (usually change "into: foo" to "into: [foo, bar]".'
                    f" If you wanted to install these paths and exclude rules are getting in your"
                    f" way, then please move this install rule before the exclusion rule that causes"
                    f" issue or, in case of built-in excludes, list the paths explicitly (without"
                    f" using patterns). Source for this issue is {self._definition_source}"
                )
            elif combined_exclude:
                msg = (
                    f"There were no matches for {pattern} in {self._search_dirs} after ignoring"
                    f" {combined_exclude} path(s) that have been excluded."
                    " If you wanted to install some of these paths, please move the install rule"
                    " before the exclusion rule or, in case of built-in excludes, list the paths explicitly"
                    f" (without using patterns). Source for this issue is {self._definition_source}."
                )
            elif already_installed_paths:
                msg = (
                    f"There were no matches for {pattern} in {self._search_dirs} after ignoring"
                    f" {already_installed_paths} path(s) already been matched previously."
                    " If you wanted to install some of these paths into multiple packages,"
                    f" please tweak the definition that installed them to install them into"
                    f' multiple packages (usually change "into: foo" to "into: [foo, bar]".'
                    f" Source for this issue is {self._definition_source}"
                )
            else:
                # TODO: Try harder to find the match and point out possible typos
                msg = (
                    f"There were no matches for {pattern} in {self._search_dirs} (definition:"
                    f" {self._definition_source})."
                )
            raise NoMatchForInstallPatternError(
                msg,
                pattern,
                self._search_dirs,
                self._definition_source,
            )
        return matched

    def _install_matches(
        self,
        path_matcher: SourcePathMatcher,
        matches: List[PathMatch],
        dest_path: Union[str, Callable[[PathMatch], str]],
        dest_path_is_format: bool,
        install_context: "BinaryPackageInstallRuleContext",
        condition_context: ConditionContext,
    ) -> None:
        if (
            self._condition is not None
            and not self._condition.evaluate(condition_context)
        ) or not install_context.binary_package.should_be_acted_on:
            # Rule is disabled; skip all its actions - also allow empty matches
            # for this particular case.
            return

        if not matches:
            raise ValueError("matches must not be empty")

        fs_root = install_context.fs_root
        for match, dest in _resolve_matches(
            matches,
            dest_path,
            dest_path_is_format,
            install_context,
        ):
            if dest.endswith("/"):
                raise ValueError(
                    f'Provided destination for "{match.path}" ended with "/" ("{dest}"), which it must not!'
                )
            dir_part, basename = os.path.split(dest)
            dir_path, missing_parts = fs_root.attempt_lookup(dir_part)
            for missing_part in missing_parts:
                # We do not associate these with the FS path.  First off,
                # it is complicated to do in most cases (indeed, debhelper
                # does not preserve these directories either) and secondly,
                # it is "only" mtime and mode - mostly irrelevant as the
                # directory is 99.9% likely to be 0755 (we are talking
                # directories like "/usr", "/usr/share").
                dir_path = dir_path.mkdir(missing_part)

            existing_path = dir_path.get(basename)

            if stat.S_ISDIR(match.stat_cache.st_mode):
                if existing_path is not None and not existing_path.is_dir:
                    existing_path.unlink()
                    existing_path = None
                current_dir = existing_path

                if current_dir is None:
                    current_dir = dir_path.mkdir(basename, fs_path=match.fs_path)
                self._install_dir_recursively(path_matcher, current_dir, match)
            else:
                if existing_path is not None and existing_path.is_dir:
                    _error(
                        f"Cannot install {match.path} ({match.fs_path}) as {dest}. That path already exist and"
                        f" is a directory.  This error was triggered via {self._definition_source}."
                    )
                dir_path.insert_file_from_fs_path(
                    basename,
                    match.fs_path,
                    follow_symlinks=False,
                    require_copy=False,
                    use_fs_path_mode=True,
                )

    def _install_dir_recursively(
        self,
        path_matcher: SourcePathMatcher,
        parent_dir: FSPath,
        match: PathMatch,
    ) -> None:
        stack = [
            (parent_dir, e)
            for e in fs_scan_dir(match.fs_path)
            if not path_matcher.is_reserved(e)
        ]

        while stack:
            current_dir, dir_entry = stack.pop()
            path_matcher.reserve(
                dir_entry.path,
                self._into,
                self._definition_source,
                is_exact_match=False,
            )
            new_dir = FSPathDirEntry(dir_entry, current_dir)
            if dir_entry.is_dir(follow_symlinks=False):
                stack.extend(
                    (new_dir, de)
                    for de in fs_scan_dir(dir_entry.path)
                    if not path_matcher.is_reserved(de)
                )

    def perform_install(
        self,
        path_matcher: SourcePathMatcher,
        install_contexts: Mapping[str, "BinaryPackageInstallRuleContext"],
        condition_context: ConditionContext,
    ) -> None:
        raise NotImplementedError

    @classmethod
    def install_as(
        cls,
        source: str,
        dest_path: str,
        into: FrozenSet[BinaryPackage],
        search_dirs: List[str],
        definition_source: str,
        condition: Optional[ManifestCondition],
    ) -> "InstallRule":
        return GenericInstallationRule(
            [source],
            dest_path,
            into,
            search_dirs,
            condition,
            definition_source,
            dest_path_is_format=False,
            require_single_match=True,
        )

    @classmethod
    def install_dest(
        cls,
        sources: List[str],
        dest_dir: Optional[str],
        into: FrozenSet[BinaryPackage],
        exclude: List[str],
        search_dirs: List[str],
        definition_source: str,
        condition: Optional[ManifestCondition],
    ) -> "InstallRule":
        if dest_dir is None:
            dest_dir = "{dirname}/{basename}"
        else:
            dest_dir = os.path.join(dest_dir, "{basename}")
        return GenericInstallationRule(
            sources,
            dest_dir,
            into,
            exclude,
            search_dirs,
            condition,
            definition_source,
            dest_path_is_format=True,
        )

    @classmethod
    def install_doc(
        cls,
        sources: List[str],
        into: FrozenSet[BinaryPackage],
        exclude: List[str],
        search_dirs: List[str],
        definition_source: str,
        condition: Optional[ManifestCondition],
    ) -> "InstallRule":
        cond = _BUILD_DOCS_BDO
        if condition is not None:
            cond = ManifestCondition.all_of([cond, condition])
        return GenericInstallationRule(
            sources,
            "usr/share/doc/{doc_main_package_name}/{basename}",
            into,
            exclude,
            search_dirs,
            cond,
            definition_source,
            dest_path_is_format=True,
        )

    @classmethod
    def install_examples(
        cls,
        sources: List[str],
        into: FrozenSet[BinaryPackage],
        exclude: List[str],
        search_dirs: List[str],
        definition_source: str,
        condition: Optional[ManifestCondition],
    ) -> "InstallRule":
        cond = _BUILD_DOCS_BDO
        if condition is not None:
            cond = ManifestCondition.all_of([cond, condition])
        return GenericInstallationRule(
            sources,
            "usr/share/doc/{doc_main_package_name}/examples/{basename}",
            into,
            exclude,
            search_dirs,
            cond,
            definition_source,
            dest_path_is_format=True,
        )

    @classmethod
    def exclude_paths(
        cls,
        paths: List[str],
        search_dirs: List[str],
        definition_source: str,
        condition: Optional[ManifestCondition],
    ) -> "InstallRule":
        return ExcludeRule(
            paths,
            search_dirs,
            condition,
            definition_source,
        )


class GenericInstallationRule(InstallRule):
    __slots__ = (
        "_sources",
        "_dest_path",
        "_dest_path_is_format",
        "_require_single_match",
    )

    def __init__(
        self,
        sources: List[str],
        dest_path: Union[str, Callable[[PathMatch], str]],
        into: FrozenSet[BinaryPackage],
        exclude: List[str],
        search_dirs: List[str],
        condition: Optional[ManifestCondition],
        definition_source: str,
        *,
        dest_path_is_format: Optional[bool] = None,
        require_single_match: bool = False,
    ) -> None:
        super().__init__(into, exclude, search_dirs, condition, definition_source)
        self._sources = sources
        self._dest_path = dest_path
        self._dest_path_is_format = (
            dest_path_is_format if dest_path_is_format is not None else False
        )
        self._require_single_match = require_single_match
        if self._require_single_match and len(sources) != 1:
            raise ValueError("require_single_match implies sources must have len 1")

    def perform_install(
        self,
        path_matcher: SourcePathMatcher,
        install_contexts: Mapping[str, "BinaryPackageInstallRuleContext"],
        condition_context: ConditionContext,
    ) -> None:
        for source in self._sources:
            matches = self._match_pattern(path_matcher, source, condition_context)
            if self._require_single_match and len(matches) != 1:
                _error(
                    f'The pattern "{source}" matched {len(matches)} entries, but it should matched'
                    f" exactly one item. Please tighten the pattern defined in {self._definition_source}"
                )
            for pkg in self._into:
                install_context = install_contexts[pkg.name]
                self._install_matches(
                    path_matcher,
                    matches,
                    self._dest_path,
                    self._dest_path_is_format,
                    install_context,
                    condition_context,
                )


class ExcludeRule(InstallRule):
    __slots__ = ("_paths",)

    def __init__(
        self,
        paths: List[str],
        search_dirs: List[str],
        condition: Optional[ManifestCondition],
        definition_source: str,
    ) -> None:
        super().__init__(frozenset(), search_dirs, condition, definition_source)
        self._paths = paths

    def perform_install(
        self,
        path_matcher: SourcePathMatcher,
        install_contexts: Mapping[str, "BinaryPackageInstallRuleContext"],
        condition_context: ConditionContext,
    ) -> None:
        for path in self._paths:
            self._match_pattern(path_matcher, path, condition_context)


class BuiltinExcludeRule(InstallRule):
    pass
